(*
All the examples below are recursive ML functions defined on lists.
What we want to notice is that the list data structure has a recursive
definition that leads naturally to writing recursive functions.

A list data structure has the following recursive structure,
   a list is either
       empty
   or the list has the structure
      item :: rest_of_the_list

Using a grammar, we might describe lists this way.

   List ::= '[]'                // "base" case
          | Item '::' List      // "recursive" case

When we write a recursive function whose input is a list,
we need to decide what to do in each of the above two cases,
the empty list case and the (recursive) non-empty list case.

Most recursive functions on a list have a definition that looks
something like the following. Notice how similar this ML pseudo
code is to the above grammar.

fun f [] = an_appropriate_value
  | f (x :: xs) = do_something_with_x  combine_with  (f xs)  // use "operator notation"

The above pattern is written as if "combine_with" was an operator
in ML. Sometimes we do the combine step using a function. In that
case, the pseudo code would look like this.

fun f [] = an_appropriate_value
  | f (x :: xs) = combine( do_something_with_x, (f xs) )  // use "function notation"

In words, we can say that the function gives the empty list
some appropriate value but, if the list is not empty,
   1) does something with the first item from the list,
   2) recursively processes the rest of the list,
   3) then somehow combines those two results.

In each of the following examples, make sure you see how the
example implements this pattern. What is the appropriate result
for the empty list? What is done to the item from the non-empty
list? How is that result combined with the result from recursively
processing the rest of the list? (Some of the examples will
modify this pattern a bit.)
*)


(* 1.
Write a function length with type
   fn : 'a list -> int
that calculates the length of the list.

length [] ==> 0
length [0] ==> 1
length ["cat", "dog"] ==> 2
*)
fun length [] = 0
  | length (_::xs) = 1 + length xs;


(* 2.
Write a function sum with type
   fn : int list -> int
that calculates the sum of the numbers in the list.

sum [] ==> 0
sum [1, 2, 3] ==> 6
*)
fun sum [] = 0
  | sum (x::xs) = x + sum xs;


(* 3.
Write a function squareThem with type
   fn : int list -> int list
that squares every number in the input list
and puts the squared number in the output list.

squareThem [3, 4, 5, 2] ==> [9, 16, 25, 4]
*)
fun squareThem [] = []
  | squareThem (x::xs) = x*x :: squareThem xs;
(*
squareThem [3,4,5,2] ==> 3*3 :: squareThem [4, 5, 2]
                     ==> 9 :: 4*4 :: squareThem [5, 2]
                     ==> 9 :: 16 :: 5*5 :: squareThem [2]
                     ==> 9 :: 16 :: 25 :: 2*2 :: squareThem []
                     ==> 9 :: 16 :: 25 :: 4 :: []
                     ==> 9 :: 16 :: 25 :: [4]
                     ==> 9 :: 16 :: [25, 4]
                     ==> 9 :: [16, 25, 4]
                     ==> [9, 16, 25, 4]
*)


(* 4.
Write a function squareSum with type
   fn : int list -> int
that sums up the squares of every number in the input list.
*)
fun squareSum [] = 0
  | squareSum (x::xs) = x*x + squareSum xs;
(*
squareSum [3,4,5,2] ==> 3*3 + squareSum [4, 5, 2]
                    ==> 9 + squareSum [4, 5, 2]
                    ==> 9 + 4*4 + squareSum [5, 2]
                    ==> 9 + 16 + squareSum [5, 2]
                    ==> 25 + squareSum [5, 2]
                    ==> 25 + 5*5 + squareSum [2]
                    ==> 25 + 25 + squareSum [2]
                    ==> 50 + squareSum [2]
                    ==> 50 + 2*2 :: squareSum []
                    ==> 50 + 4 :: squareSum []
                    ==> 54 :: squareSum []
                    ==> 54 + 0
                    ==> 54
*)


(* 5.
Write a function implode with type
   fn : char list -> string
that concatenates all the characters from the input list into a string.
*)
fun implode [] = ""
  | implode (c::cs) = (str c) ^ implode cs;
(*
implode [#"d", #"o", #"g", #"s"] ==> (str #"d") ^ implode [#"o", #"g", #"s"]
                                 ==> "d" ^ (str #"o") ^ implode [#"g", #"s"]
                                 ==> "d" ^ "o" ^ (str #"g") ^ implode [#"s"]
                                 ==> "d" ^ "o" ^ "g" ^ (str #"s") ^ implode []
                                 ==> "d" ^ "o" ^ "g" ^ "s" ^ ""
                                 ==> "d" ^ "o" ^ "g" ^ "s"
                                 ==> "d" ^ "o" ^ "gs"
                                 ==> "d" ^ "ogs"
*)


(* 6.
Write a function max with type
   fn : int list -> int
that finds the maximum number in the list.

Use ML's Int.max function to find the max of two numbers.

Define the max of the empty list as the ML library's
pre-defined minimum integer value, "valOf Int.minInt".
*)
fun max [] = valOf Int.minInt      (* -infinity *)
  | max [x] = x    (* this clause is NOT needed *)
  | max (x::xs) = Int.max(x, max xs);
(*
max [3,6,2,5]  ==> Int.max (3, max [6, 2, 5])
               ==> Int.max (3, Int.max (6, max [2, 5]))
               ==> Int.max (3, Int.max (6, Int.max (2, max [5])))
               ==> Int.max (3, Int.max (6, Int.max (2, Int.max (5, max []))))
               ==> Int.max (3, Int.max (6, Int.max (2, Int.max (5, minInt))))
               ==> Int.max (3, Int.max (6, Int.max (2, 5)))
               ==> Int.max (3, Int.max (6, 5))
               ==> Int.max (3, 6)
               ==> 6
*)


(* 7.
Write a function maxPairs with type
   fn : (int*int) list -> int list
where each item in the output list is the maximum of the
two elements from the corresponding pair in the input list.
*)
fun maxPairs [] = []
  | maxPairs ((x,y)::ps) = Int.max(x,y) :: maxPairs ps;


(* 8.
Write a function multPairs with type
   fn : (int*int) list -> int list
where each item in the output list is the product of the
two elements from the corresponding pair in the input list.
*)
fun multPairs [] = []
  | multPairs ((x,y)::ps) = x*y :: multPairs ps;


(* 9.
Write a function multAllPairs with type
   fn : (int*int) list -> int
where the output number is the product of all the
numbers form all the pairs from the input list.
*)
fun multAllPairs [] = 1
  | multAllPairs ((x,y)::ps) = x*y * multAllPairs ps;


(* 10.
Write a function dupList with type
  fn : 'a list -> 'a list
that duplicates into the output list every item from the
input list. So the length of the output list is twice
the length of the input list.

dupList [1, 2, 3, 4] ==> [1, 1, 2, 2, 3, 3, 4, 4]
*)
fun dupList [] = []
  | dupList (x::xs) = x :: x :: dupList xs;
(* or *)
fun dupList [] = []
  | dupList (x::xs) = [x, x] @ dupList xs;


(* 11.
Write a function pairUp with the type
   fn : 'a list -> ('a * 'a) list
that pairs up every two successive items from the input list
into a tuple for the output list. A one-item list has nothing
to pair the one item with so you should return an empty list.
(What other possible result can you think of for the one-item case?)

pairUp [1,2,3,4] ==> [(1,2), (3,4)]
pairUp [1,2,3,4,5] ==> [(1,2), (3,4)]
pairUp [1] ==> []
pairUp [] ==> []
*)
fun pairUp [] = []   (* base case of an empty list *)
  | pairUp [x] = []  (* base case for a one-item list *)
  | pairUp (x::y::ys) = (x,y) :: pairUp ys;  (* recurse two items at a time *)
(*
Notice why we need BOTH base cases. If the input list has
even length, we end up landing on the empty-list base case.
But if the input list has odd length, then we end up landing
on the one-item base case.

Take the above definition and remove from it the one-tem base
case. Then test the new definition to see that it still works
on even length lists, but fails on odd length lists.
*)


(* 12.
Write a function zipUp with the type
   fn : 'a list * 'b list -> ('a * 'b) list
that takes in a pair of lists and outputs a list of pairs.
The function should pair up corresponding elements from the
two input lists. The length of the output list is the minimum
of the lengths of the two input lists.

This example is a different from the previous ones since it
recurses on two lists at the same time. If either list is
empty, the result is the empty list. If neither list is empty,
then pair up the given elements into the output list.
*)
fun zipUp ([], _) = []   (* base case for the first list *)
  | zipUp (_, []) = []   (* base case for the second list *)
  | zipUp (x::xs, y::ys) = (x,y) :: zipUp (xs, ys);
(*
zipUp ([1,2,3], [10,11,12,13,14]) ==> (1,10) :: zipUp ([2,3], [11,12,13,14])
                                  ==> (1,10) :: (2,11) :: zipUp ([3], [12,13,14])
                                  ==> (1,10) :: (2,11) :: (3,12) :: zipUp ([], [13,14])
                                  ==> (1,10) :: (2,11) :: (3,12) :: []
                                  ==> [(1,10), (2,11), (3,12)]
*)


(* 13.
Write a function take with type
   fn : int * 'a list -> 'a list
where the output list is the first n items from
the input list, with n being the int parameter.

This example is a different from the previous ones since
it recurses simultaneously on both a list and a number.

take (3, [1, 2, 3, 4, 5, 6]) ==> [1, 2, 3]
take (5, [1, 2, 3]) ==> [1, 2, 3]
take (0, [1, 2, 3]) ==> []
*)
fun take (_, []) = []  (* base case for the list *)
  | take (0, _)  = []  (* base case for the number *)
  | take (n, x::xs) = x :: take (n - 1, xs);
(*
take (3, [1,2,3,4,5,6]) ==> 1 :: take (2, [2,3,4,5,6])
                        ==> 1 :: 2 :: take (1, [3,4,5,6])
                        ==> 1 :: 2 :: 3 :: take (0, [4,5,6])
                        ==> 1 :: 2 :: 3 :: []
                        ==> [1,2,3]

take (5, [1,2,3]) ==> ==> 1 :: take (4, [2,3])
                      ==> 1 :: 2 :: take (3, [3])
                      ==> 1 :: 2 :: 3 :: take (2, [])
                      ==> 1 :: 2 :: 3 :: []
                      ==> [1,2,3]
*)


(* 14.
Write a function drop with type
   fn : int * 'a list -> 'a list
where the output list is the input list with its first
n items removed (dropped), with n being the int parameter.

This example, like the previous one, recurses simultaneously
on both a list and a number.

drop (3, [1, 2, 3,4, 5, 6]) ==> [4, 5, 6]
drop (5, [1, 2, 3]) ==> []
drop (0, [1, 2, 3]) ==> [1, 2, 3]
*)
fun drop (_, []) = []   (* base case for the list *)
  | drop (0, xs)  = xs  (* base case for the number *)
  | drop (n, _::xs) = drop (n - 1, xs);
(*
drop (3, [1,2,3,4,5,6]) ==> drop (2, [2, 3, 4, 5, 6])
                        ==> drop (1, [3, 4, 5, 6])
                        ==> drop (0, [4, 5, 6])
                        ==> [4, 5, 6]

drop (5, [1,2,3]) ==> ==> drop (4, [2, 3])
                      ==> drop (3, [3])
                      ==> drop (2, [])
                      ==> []
*)


(* 15.
Write a function flatten with this type
   fn : 'a list list -> 'a list
which takes as input a list of lists and "flattens" it into
just a list. The output list should contain all the items
from all the sub-lists of the input list in the order that
they were in in the input list.

This example is a tricky because it recurses on a list of lists.
This example recurses on the given list and it simultaneously recurses
on each sub-list of the given list. We need a base case for the list
itself, and we need a base case for each sub-list.

flatten [[1,2,3], [4,5], [6,7,8], [], [9]] ==> [1,2,3,4,5,6,7,8,9]
*)
fun flatten [] = []                           (* empty list base case *)
  | flatten ([]::listList) = flatten listList (* empty sub-list base case *)
  | flatten ((x::xs)::listList) = x :: flatten (xs :: listList)
(*
flatten [[1,2,3],[4,5],[6,7,8],[],[9]] ==> 1::flatten [[2,3],[4,5],[6,7,8],[],[9]]
                                       ==> 1::2::flatten [[3],[4,5],[6,7,8],[],[9]]
                                       ==> 1::2::3::flatten [[],[4,5],[6,7,8],[],[9]]
                                       ==> 1::2::3::flatten [[4,5],[6,7,8],[],[9]]
                                       ==> 1::2::3::4::flatten [[5],[6,7,8],[],[9]]
                                       ==> 1::2::3::4::5::flatten [[],[6,7,8],[],[9]]
                                       ==> 1::2::3::4::5::flatten [[6,7,8],[],[9]]
                                       ==>
*)


(* 16.
Write a function sumOfLengths with this type
   fn : 'a list list -> int
which takes as input a list of lists and computes the
"total" length of the list. So the result is the sum
of the lengths of the sub-lists.

sumOfLengths [[1,2,3], [4,5], [6,7,8], [], [9]] ==> 9
*)

(* Here are three solutions. Which one is "best"? Why? *)
fun sumOfLengths listList = length (flatten listList);

fun sumOfLengths [] = 0
  | sumOfLengths (xs::listList) = (length xs) + (sumOfLengths listList);

fun sumOfLengths [] = 0
  | sumOfLengths ([]::listList) = sumOfLengths listList
  | sumOfLengths ((_::xs)::listList) = 1 + sumOfLengths (xs :: listList);


(* 17.
Write a function sumOfLists with this type
   fn : int list list -> int
which takes as input a list of integer lists and computes the
total sum of all the sub-lists.

sumOfLists [[1,2,3], [4,5], [6,7,8], [], [9]] ==> 45
*)

(* Here are three solutions. *)
fun sumOfLists listList = sum (flatten listList);

fun sumOfLists [] = 0
  | sumOfLists (xs::listList) = (sum xs) + (sumOfLists listList);

fun sumOfLists [] = 0
  | sumOfLists ([]::listList) = sumOfLists listList
  | sumOfLists ((x::xs)::listList) = x + sumOfLists (xs :: listList);


(* 18.
Write a function repeatList with type
   fn : int * 'a list -> 'a list
where the output list is the input list repeated n times
with n the integer parameter. The length of the output
list will be n * length (input list).

This example is different from the others because the
recursion is not really on the list. The recursion is on
the number. This is "counting recursion", more like a
loop in Java.

repeatList (3, ["cat","dog","car"]) ==> ["cat","dog","car","cat","dog","car","cat","dog","car"]
repeatList (5, [0,1,2,3,4]) ==> [0,1,2,3,4,0,1,2,3,4,0,1,2,3,4,0,1,2,3,4,0,1,2,3,4]
*)
fun repeatList (_, []) = []  (* this clause is not really needed *)
  | repeatList (0, _)  = []  (* the base case is on the number *)
  | repeatList (n, xs) = xs @ repeatList (n-1, xs); (* the recursion isn't on the list! *)


(* 19.
Write a function repeat with type
   fn : int * 'a -> 'a list
where the output list is the input item repeated n times
with n the integer parameter.

The recursion is again on the number. This is "counting recursion",
more like a loop in Java.

Explain why the following two functions are not the same.
  repeat (3, ["cat","dog","car"])
  repeatList (3, ["cat","dog","car"])
*)

fun repeat (0, _)  = []  (* the base case is on the number *)
  | repeat (n, x) = x :: repeat (n-1, x);



(* 20.
Write a function indexing with type
   fn : 'a list -> (int * 'a) list
where each element from the input list is paired up with
its index in the input list (with the first index being 0).

indexing ["cat","dog","car","fun"] ==> [(0,"cat"),(1,"dog"),(2,"car"),(3,"fun")]
*)

(* This does not work! Why? *)
fun indexing [] = []
  | indexing (x::xs) = (0, x) :: indexing xs;

(* We need a "helper function. *)
fun indexingHelper (_, []) = []
  | indexingHelper (n, x::xs) = (n, x) :: indexingHelper (n+1, xs);

fun indexing xs = indexingHelper (0, xs);


(* Rewrite indexing with its helper function hidden inside. *)
fun indexing xs =
  let
    fun indexingHelper (_, []) = []
      | indexingHelper (n, x::xs) = (n, x) :: indexingHelper (n+1, xs)
  in
    indexingHelper (0, xs)
  end;


(* 21.
Write a function allOrderedPairs with type
   fn : int*int -> (int*int) list
that generates all possible ordered pairs (a,b)
with 1 <= a <= n and a <= b <= m
and where n and m are the function parameters.

This is a double counting recursion, kind of like
nested for-loops in Java. It needs a nested (recursive)
helper function. It is quite a bit more complicated to
understand than nested for-loops.
*)
fun allOrderedPairs (_, 0) = [] (* this is not really needed *)
  | allOrderedPairs (0, _) = [] (* base case for iterating on n *)
  | allOrderedPairs (n, m) =
     let
        fun rowN 0 = []   (* base case for iterating on m *)
          | rowN m = (n, m) :: rowN (m - 1)
     in
        (rowN m) @ (allOrderedPairs (n - 1, m))
     end;
